#include "psdraw.h"
#include <map>

/*
  3d coordinate system
 y^ z
  |/
  +->x

  2d coordinate system
  +-> x
  |
 yv
  
 All units: 72 Coords to an inch
*/

// see also elttypes.h and graphcover.cpp
class DrawElt
  {
public:
    Coord Z;
    DrawElt(Coord inZ) : Z(inZ) { }
    virtual ~DrawElt() { }
    virtual void Render(DrawTo *)=0;
  };

inline bool compare(DrawElt *a, DrawElt *b)
  {
    return a->Z > b->Z;
  }

inline void Project(Coord in[3], Coord out[2])
  {
    out[0]=in[0];
    out[1]=up*(in[1]+in[2]);
  }

// in graphcover.cpp:
// Do we want scenes to remove their elements as they render them?
// Choose between a priority queue and a vector which is sorted at render time.
// Only makes a difference if want to Render multiple times (where you probably want a priority queue).
class Scene
  {
private:
    typedef instl vector<DrawElt *> containertype;
    typedef containertype::iterator iter;
    containertype allrenderable;
    DrawTo *renderto;
public:
    Scene(DrawTo *dto) : renderto(dto) { }
    ~Scene() 
      { for (iter i=allrenderable.begin(); i!=allrenderable.end(); ++i)
          delete (*i); }
    Scene & operator <<(DrawElt *toadd) 
      { allrenderable.push_back(toadd); return *this; }
    void Render(); // renders all DrawElt's in order of decreasing Z
    DrawTo *GetDrawTo() { return renderto; }
  };

// in graphcover.cpp:
class Settings
  {
public:
    Coord LineRNormal,
          LineRThick,
          DiscRNormal,
          DiscRThick;
    bool DiscRThickSet, LineRThickSet;
    bool DrawBaseLabels, DrawCoverLabels;
    int detail;
    bool yisup;
    Coord LoopSize,
          LoopAspect,
          MultiEdgeWidth;
    Coord CoverYStretch;
    Coord CoverRandomDisplace;
    bool  CoverEnableCurves;

    Settings();
    bool Setting(string name, Coord value); // returns true if setting is recognized
  };

extern Settings setting;
extern Coord epsilon;

class Graph;

// in vertex.cpp:
class Vertex
  {
public:
    Coord P[3];
    string label;
    Coord HaloDepth;
    Coord TextDepth;

    virtual ~Vertex() { }
    void Set(Coord inP[3], string labeltext);
    virtual void Render(Scene &);
    void UpdateBounds(Graph &graph, DrawTo *dt);
    void Translate(Coord t[3]);
  };

struct GraphPoint
  {
    Coord p[3];
  };
typedef instl list<GraphPoint> EdgeSpline;
typedef EdgeSpline::iterator EdgeSplineIter;

// in edge.cpp:
class Edge
  {
protected:
    void GenerateSpline(int degree, Coord Control[][3]);
    void UpdateVertex(int whichv, Coord Z1, Coord Z2, Graph &graph);

public:
    int v[2];
    typedef instl list<EdgeSpline> arclist;
    typedef arclist::iterator arciter;
    arclist arcs;

      // Set modifies neighboring vertices to have correct
      // HaloDepth and TextDepth, by calling ModifyNeighbors
    void Set(int v1, int v2, int multiplicity, Graph &graph);
    void ModifyNeighbors(Graph &graph);
    void IncreaseMultiplicity(int multiplicity, Graph &graph);
    void UpdateBounds(Graph &graph);
    void Translate(Coord t[3]);
    void Render(Scene &);
  };

// in graph.cpp
class Graph
  {
public:
    Coord Center[3];
    Coord Bounds[2][3]; // Bounds[0]=min Bounds[1]=max
    instl vector<Edge> e;

    virtual int AddEdge(int v1, int v2, int multiplicity=1);
    int FindEdge(int v1, int v2);

    virtual Vertex &Vert(int v) = 0;
    virtual int NumVert() = 0;

    int FindVertex(string label);
    int FindVertexSubstring(string label);

    void UpdateBounds(DrawTo *dt);
    void Translate(Coord t[3]);

    void Render(Scene &);
  };

// in graph.cpp
class GraphBase : public Graph
  {
public:
    instl vector<Vertex> v;

    Vertex &Vert(int i);
    int NumVert();

    int AddVertex(Coord P[3], string label);
  };

class CoverVertex : public Vertex
  {
public:
    int basevertex;
    virtual void Render(Scene &);
  };

typedef instl vector<int> VertexFiber;  
// index of cover vertex over each base vertex

class EdgeFiber
  {
public:
    struct PerEdge
      {
        int othervertex;
        int edgenum;
        PerEdge() { }
        PerEdge(int aOthervertex, int aEdgenum)
          { othervertex=aOthervertex; edgenum=aEdgenum; }
      };

    typedef instl list<PerEdge> edgelist;
    typedef edgelist::iterator  edgeiter;

    static int deg;
    struct PerVertex
      {
        edgelist Edges;
        PerVertex() { }
      };
    typedef instl map<int, PerVertex> vertlist;
    typedef vertlist::iterator        vertiter;

    vertlist  mV;
  };


// in graph.cpp
class GraphCover : public Graph
  {
public:
    instl vector<CoverVertex> v;
    //instl vector<Edge>        e;
    instl vector<VertexFiber> vf;
    instl vector<EdgeFiber>   ef;
    GraphBase *base;
    Coord yspacing;

    GraphCover() { base=0; }

    Vertex &Vert(int i);
    int NumVert();

    int AddEdge(int v1, int v2, int multiplicity=1);

    void Initialize(GraphBase *newbase, Coord newyspacing);
    bool WasInitialized() { return base!=0; }
    int AddVertex(string label);
    void Rearrange();
  };

// In search.cpp:
bool AssignCoverEdge(GraphCover *cover, int baseedge);

